@@ -17,7 +17,7 @@ require_relative "bundler/build_metadata"
 # Bundler provides a consistent environment for Ruby projects by
 # tracking and installing the exact gems and versions that are needed.
 #
-# Since Ruby 2.6, Bundler is a part of Ruby's standard library.
+# Bundler is a part of Ruby's standard library.
 #
 # Bundler is used by creating _gemfiles_ listing all the project dependencies
 # and (optionally) their versions and then using
@@ -39,6 +39,8 @@ module Bundler
   environment_preserver.replace_with_backup
   SUDO_MUTEX = Thread::Mutex.new
 
+  autoload :Checksum,               File.expand_path("bundler/checksum", __dir__)
+  autoload :CIDetector,             File.expand_path("bundler/ci_detector", __dir__)
   autoload :Definition,             File.expand_path("bundler/definition", __dir__)
   autoload :Dependency,             File.expand_path("bundler/dependency", __dir__)
   autoload :Deprecate,              File.expand_path("bundler/deprecate", __dir__)
@@ -98,9 +100,7 @@ module Bundler
     end
 
     def create_bundle_path
-      SharedHelpers.filesystem_access(bundle_path.to_s) do |p|
-        mkdir_p(p)
-      end unless bundle_path.exist?
+      mkdir_p(bundle_path) unless bundle_path.exist?
 
       @bundle_path = bundle_path.realpath
     rescue Errno::EEXIST
@@ -117,7 +117,7 @@ module Bundler
       @bin_path ||= begin
         path = settings[:bin] || "bin"
         path = Pathname.new(path).expand_path(root).expand_path
-        SharedHelpers.filesystem_access(path) {|p| FileUtils.mkdir_p(p) }
+        mkdir_p(path)
         path
       end
     end
@@ -192,7 +192,7 @@ module Bundler
     end
 
     def environment
-      SharedHelpers.major_deprecation 2, "Bundler.environment has been removed in favor of Bundler.load", :print_caller_location => true
+      SharedHelpers.major_deprecation 2, "Bundler.environment has been removed in favor of Bundler.load", print_caller_location: true
       load
     end
 
@@ -210,9 +210,10 @@ module Bundler
     end
 
     def frozen_bundle?
-      frozen = settings[:deployment]
-      frozen ||= settings[:frozen]
-      frozen
+      frozen = settings[:frozen]
+      return frozen unless frozen.nil?
+
+      settings[:deployment]
     end
 
     def locked_gems
@@ -329,14 +330,6 @@ module Bundler
 
     def rm_rf(path)
       FileUtils.remove_entry_secure(path) if path && File.exist?(path)
-    rescue ArgumentError
-      message = <<EOF
-It is a security vulnerability to allow your home directory to be world-writable, and bundler cannot continue.
-You should probably consider fixing this issue by running `chmod o-w ~` on *nix.
-Please refer to https://ruby-doc.org/stdlib-3.1.2/libdoc/fileutils/rdoc/FileUtils.html#method-c-remove_entry_secure for details.
-EOF
-      File.world_writable?(path) ? Bundler.ui.warn(message) : raise
-      raise PathError, "Please fix the world-writable issue with your #{path} directory"
     end
 
     def settings
@@ -352,13 +345,13 @@ EOF
 
     # @deprecated Use `unbundled_env` instead
     def clean_env
-      Bundler::SharedHelpers.major_deprecation(
-        2,
+      message =
         "`Bundler.clean_env` has been deprecated in favor of `Bundler.unbundled_env`. " \
-        "If you instead want the environment before bundler was originally loaded, use `Bundler.original_env`",
-        :print_caller_location => true
-      )
-
+        "If you instead want the environment before bundler was originally loaded, use `Bundler.original_env`"
+      removed_message =
+        "`Bundler.clean_env` has been removed in favor of `Bundler.unbundled_env`. " \
+        "If you instead want the environment before bundler was originally loaded, use `Bundler.original_env`"
+      Bundler::SharedHelpers.major_deprecation(2, message, removed_message: removed_message, print_caller_location: true)
       unbundled_env
     end
 
@@ -395,13 +388,13 @@ EOF
 
     # @deprecated Use `with_unbundled_env` instead
     def with_clean_env
-      Bundler::SharedHelpers.major_deprecation(
-        2,
+      message =
         "`Bundler.with_clean_env` has been deprecated in favor of `Bundler.with_unbundled_env`. " \
-        "If you instead want the environment before bundler was originally loaded, use `Bundler.with_original_env`",
-        :print_caller_location => true
-      )
-
+        "If you instead want the environment before bundler was originally loaded, use `Bundler.with_original_env`"
+      removed_message =
+        "`Bundler.with_clean_env` has been removed in favor of `Bundler.with_unbundled_env`. " \
+        "If you instead want the environment before bundler was originally loaded, use `Bundler.with_original_env`"
+      Bundler::SharedHelpers.major_deprecation(2, message, removed_message: removed_message, print_caller_location: true)
       with_env(unbundled_env) { yield }
     end
 
@@ -417,13 +410,13 @@ EOF
 
     # @deprecated Use `unbundled_system` instead
     def clean_system(*args)
-      Bundler::SharedHelpers.major_deprecation(
-        2,
+      message =
         "`Bundler.clean_system` has been deprecated in favor of `Bundler.unbundled_system`. " \
-        "If you instead want to run the command in the environment before bundler was originally loaded, use `Bundler.original_system`",
-        :print_caller_location => true
-      )
-
+        "If you instead want to run the command in the environment before bundler was originally loaded, use `Bundler.original_system`"
+      removed_message =
+        "`Bundler.clean_system` has been removed in favor of `Bundler.unbundled_system`. " \
+        "If you instead want to run the command in the environment before bundler was originally loaded, use `Bundler.original_system`"
+      Bundler::SharedHelpers.major_deprecation(2, message, removed_message: removed_message, print_caller_location: true)
       with_env(unbundled_env) { Kernel.system(*args) }
     end
 
@@ -439,13 +432,13 @@ EOF
 
     # @deprecated Use `unbundled_exec` instead
     def clean_exec(*args)
-      Bundler::SharedHelpers.major_deprecation(
-        2,
+      message =
         "`Bundler.clean_exec` has been deprecated in favor of `Bundler.unbundled_exec`. " \
-        "If you instead want to exec to a command in the environment before bundler was originally loaded, use `Bundler.original_exec`",
-        :print_caller_location => true
-      )
-
+        "If you instead want to exec to a command in the environment before bundler was originally loaded, use `Bundler.original_exec`"
+      removed_message =
+        "`Bundler.clean_exec` has been removed in favor of `Bundler.unbundled_exec`. " \
+        "If you instead want to exec to a command in the environment before bundler was originally loaded, use `Bundler.original_exec`"
+      Bundler::SharedHelpers.major_deprecation(2, message, removed_message: removed_message, print_caller_location: true)
       with_env(unbundled_env) { Kernel.exec(*args) }
     end
 
@@ -488,7 +481,7 @@ EOF
       configured_bundle_path.use_system_gems?
     end
 
-    def mkdir_p(path, options = {})
+    def mkdir_p(path)
       SharedHelpers.filesystem_access(path, :write) do |p|
         FileUtils.mkdir_p(p)
       end
@@ -514,7 +507,16 @@ EOF
     end
 
     def safe_load_marshal(data)
-      load_marshal(data, :marshal_proc => SafeMarshal.proc)
+      if Gem.respond_to?(:load_safe_marshal)
+        Gem.load_safe_marshal
+        begin
+          Gem::SafeMarshal.safe_load(data)
+        rescue Gem::SafeMarshal::Reader::Error, Gem::SafeMarshal::Visitors::ToRuby::Error => e
+          raise MarshalError, "#{e.class}: #{e.message}"
+        end
+      else
+        load_marshal(data, marshal_proc: SafeMarshal.proc)
+      end
     end
 
     def load_gemspec(file, validate = false)
@@ -523,7 +525,7 @@ EOF
       @gemspec_cache[key] ||= load_gemspec_uncached(file, validate)
       # Protect against caching side-effected gemspecs by returning a
       # new instance each time.
-      @gemspec_cache[key].dup if @gemspec_cache[key]
+      @gemspec_cache[key]&.dup
     end
 
     def load_gemspec_uncached(file, validate = false)
